/*
 *  2 Kernel Monte a.k.a. Linux loading Linux on x86
 *
 *  Erik Arjan Hendriks <hendriks@scyld.com>
 *  Copyright (C) 2000 Scyld Computing Corporation
 *
 *  Modified for Linux 2.4 and misc changes:
 *
 *    Christoph Hellwig <hch@caldera.de>
 *    Copyright (C) 2001 Caldera International, Inc
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  $Id: kmonte.c,v 1.18 2000/10/30 16:53:50 hendriks Exp $
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/pagemap.h>
#include <linux/malloc.h>
#include <linux/pci.h>
#include <asm/page.h>
#include <asm/uaccess.h>
#include <asm/unistd.h>
#include <asm/pgtable.h>
#include <asm/io.h>

#include "monte.h"		/* For user interface stuff.... */

MODULE_AUTHOR("Erik Arjan Hendriks <hendriks@scyld.com>");
MODULE_DESCRIPTION("Two Kernel Monte: Loads new Linux kernels from Linux.");
    

/*
 * Monte memory management
 *
 *
 * TODO: explain in excruciating detail wtf is going on here.
 */
static struct monte_reloc_page_t *	monte_pg_list = 0;

struct monte_reloc_t {
	unsigned long dest;
	unsigned long src;
};

#define MONTE_RELOCS_PER_PAGE ((PAGE_SIZE-(sizeof(long)*2))/sizeof(struct monte_reloc_t))
struct monte_reloc_page_t {
	struct monte_reloc_t relocs[MONTE_RELOCS_PER_PAGE];
	struct monte_reloc_page_t *next;
	unsigned long _pad;
};


static void
monte_pg_list_free(void)
{
	struct monte_reloc_page_t *tmp;
	while (monte_pg_list) {
		tmp = monte_pg_list->next;
		free_page((unsigned long) monte_pg_list);
		monte_pg_list = tmp;
	}
}

/* This function borrowed from the /proc file system. */
static unsigned long get_phys_addr(unsigned long ptr)
{
	pgd_t *page_dir;
	pmd_t *page_middle;
	pte_t pte;

	if (!current->mm || ptr >= TASK_SIZE)
		return 0;
	/* Check for NULL pgd .. shouldn't happen! */
	if (!current->mm->pgd) {
		printk("get_phys_addr: pid %d has NULL pgd!\n",
		       current->pid);
		return 0;
	}

	page_dir = pgd_offset(current->mm, ptr);
	if (pgd_none(*page_dir))
		return 0;
	if (pgd_bad(*page_dir)) {
		printk("bad page directory entry %08lx\n",
		       pgd_val(*page_dir));
		pgd_clear(page_dir);
		return 0;
	}
	page_middle = pmd_offset(page_dir, ptr);
	if (pmd_none(*page_middle))
		return 0;
	if (pmd_bad(*page_middle)) {
		printk("bad page middle entry %08lx\n",
		       pmd_val(*page_middle));
		pmd_clear(page_middle);
		return 0;
	}
	pte = *pte_offset(page_middle, ptr);
	if (!pte_present(pte))
		return 0;
	return (pte_val(pte) & PAGE_MASK) + (ptr & ~PAGE_MASK);
}

static int monte_setup_page_list(struct monte_region_t *regions, int nregions)
{
	int i, npages, pg_idx;
	unsigned long j;
	struct monte_reloc_page_t *pg;

	npages = 0;
	for (i = 0; i < nregions; i++)
		npages += (regions[i].size + PAGE_SIZE - 1) >> PAGE_SHIFT;
	if (npages == 0)
		return -EINVAL;

	/* Allocate all the pages we're going to need. */
	monte_pg_list = 0;
	for (i = 0; i < npages; i += MONTE_RELOCS_PER_PAGE) {
		/* grab a new page and make entries on this page. */
		if (!
		    (pg =
		     (struct monte_reloc_page_t *)
		     get_free_page(GFP_KERNEL))) return -ENOMEM;
		memset(pg, 0, sizeof(*pg));
		pg->next = monte_pg_list;
		monte_pg_list = pg;
	}

	/* Load up the page list.  Here we go through the regions that the
	 * user provided us and find the physical addresses of those
	 * pages.  XXX Should we also go through and lock these pages at
	 * this time ? */
	pg = monte_pg_list;
	pg_idx = 0;
	for (i = 0; i < nregions; i++) {
		for (j = 0; j < regions[i].size; j += PAGE_SIZE) {
			pg->relocs[pg_idx].dest =
			    ((unsigned long) regions[i].destaddr) + j;
			pg->relocs[pg_idx].src =
			    get_phys_addr(((unsigned long) regions[i].addr)
					  + j);
			if (pg->relocs[pg_idx].src == 0) {
				printk
				    ("monte: page not present in page list load!\n");
				return -ENOMEM;
			}
			pg_idx++;
			if (pg_idx == MONTE_RELOCS_PER_PAGE) {
				pg_idx = 0;
				pg = pg->next;
			}
		}
	}
	return 0;
}

static int monte_check_page_list(void)
{
	int pg1i, pg2i;
	void *tmp_page, *src;
	unsigned long tmp;
	struct monte_reloc_page_t *pg1, *pg2, *last;

	if (!(tmp_page = (void *) get_free_page(GFP_KERNEL)))
		return -ENOMEM;

	printk("monte: sorting pages: ");
	pg1 = monte_pg_list;
	pg1i = 0;
	while (pg1) {
		if (!pg1->relocs[pg1i].dest)
			goto __end;
		
		/*
		 * check if there are any pages which will be placed on
		 * this page in the end.
		 */
		pg2 = pg1;
		pg2i = pg1i + 1;
		if (pg2i == MONTE_RELOCS_PER_PAGE) {
			pg2 = pg2->next;
			pg2i = 0;
		}
		while (pg2) {
			/* We have a winner, swap these two pages. */
			if (pg2->relocs[pg2i].src == pg1->relocs[pg1i].dest) {
				/*
				 * Swap page contents
				 */
				memcpy(tmp_page,
				       phys_to_virt(pg1->relocs[pg1i].src),
				       PAGE_SIZE);
				memcpy(phys_to_virt(pg1->relocs[pg1i].src),
				       phys_to_virt(pg2->relocs[pg2i].src),
				       PAGE_SIZE);
				memcpy(phys_to_virt(pg2->relocs[pg2i].src),
				       tmp_page,
				       PAGE_SIZE);
				/*
				 * Swap "src" pointers
				 */
				tmp = pg1->relocs[pg1i].src;
				pg1->relocs[pg1i].src = pg2->relocs[pg2i].src;
				pg2->relocs[pg2i].src = tmp;
				goto __end;
			}
			
			pg2i++;
			if (pg2i == MONTE_RELOCS_PER_PAGE) {
				pg2 = pg2->next;
				pg2i = 0;
			}
		}
		
		/*
		 * Also check on the pages used as indirect pages.
		 */
		last = 0;
		for (pg2 = monte_pg_list; pg2; pg2 = pg2->next) {
			if (virt_to_phys(pg2) != pg1->relocs[pg1i].dest) {
				last = pg2;
				continue;
			}
			src = phys_to_virt(pg1->relocs[pg1i].src);
				
			/*
			 * Swap page contents
			 */
			memcpy(tmp_page, src, PAGE_SIZE);
			memcpy(src, pg2, PAGE_SIZE);
			memcpy(pg2, tmp_page, PAGE_SIZE);
				
			/*
			 * Swap the pointers
			 */
			if (last)
				last->next = src;
			else
				monte_pg_list = src;
			if (pg1 == pg2)
				pg1 = src;
			pg1->relocs[pg1i].src =
				virt_to_phys(pg2);
			break;
		}
__end:
		pg1i++;
		if (pg1i == MONTE_RELOCS_PER_PAGE) {
			pg1 = pg1->next;
			pg1i = 0;
		}
	}
	printk("done\n");
	free_page((unsigned long) tmp_page);
	return 0;
}

static void monte_make_phys(void)
{
	struct monte_reloc_page_t *pg;
	/* Convert all the pointers in our relocation list to physical
	 * addresses.  Most of them already are.  We just need to fix the
	 * "next" pointers. */
	pg = monte_pg_list;
	monte_pg_list = (void *) virt_to_phys(monte_pg_list);
	while (pg->next) {
		pg->next = (void *) virt_to_phys(pg->next);
		pg = phys_to_virt((unsigned long) pg->next);
	}
}

/*--------------------------------------------------------------------
 *  Syscall interface
 *------------------------------------------------------------------*/
static int monte_restart(unsigned long entry_addr, unsigned long flags);
int (*real_reboot) (int, int, int, void *);

asmlinkage int sys_monte(int magic1, int magic2, int cmd, void *arg)
{
	int err;
	struct monte_param_t param;
	struct monte_region_t *regions = 0;
	static DECLARE_MUTEX(monte_sem);

	MOD_INC_USE_COUNT;
	if (magic1 != MONTE_MAGIC_1 || magic2 != MONTE_MAGIC_2) {
		err = real_reboot(magic1, magic2, cmd, arg);
		MOD_DEC_USE_COUNT;
		return err;
	}
	if (!capable(CAP_SYS_BOOT)) {
		MOD_DEC_USE_COUNT;
		return -EPERM;
	}

	down(&monte_sem);	/* one process only please... */
	/* Get the user's parameters */
	if (copy_from_user(&param, arg, sizeof(param))) {
		err = -EFAULT;
		goto out;
	}
	if (!
	    (regions =
	     kmalloc(sizeof(struct monte_region_t) * param.nregions,
		     GFP_KERNEL))) {
		err = -ENOMEM;
		goto out;
	}
	if (copy_from_user
	    (regions, param.regions, sizeof(*regions) * param.nregions)) {
		err = -EFAULT;
		goto out;
	}
	down_write(&current->mm->mmap_sem);
	if ((err = monte_setup_page_list(regions, param.nregions)))
		goto out1;
	if ((err = monte_check_page_list()))
		goto out1;


	if (param.flags & MONTE_NOT_REALLY) {
		err = 0;
		goto out1;
	}

	printk("monte: restarting system in 3 seconds...\n");
	{
		unsigned long now = jiffies;
		while (jiffies < (now + 3 * HZ))
			schedule();
	}
	err = monte_restart(param.entrypoint, param.flags);
	printk("monte: failure (errno = %d)\n", -err);

      out1:
	up_write(&current->mm->mmap_sem);
      out:
	if (regions)
		kfree(regions);
	monte_pg_list_free();
	up(&monte_sem);
	MOD_DEC_USE_COUNT;
	return err;
}


/*-------------------------------------------------------------------------
 * Machine restart code
 *-----------------------------------------------------------------------*/
static unsigned long long real_mode_gdt_entries[] = {
	0x0000000000000000ULL,	/* 00h: Null descriptor */
	0x0000000000000000ULL,	/* 08h: Unused... */
	0x00cf9a000000ffffULL,	/* 10h: 32-bit 4GB code at 0x00000000 */
	0x00cf92000000ffffULL,	/* 18h: 32-bit 4GB data at 0x00000000 */
	0x00009a000000ffffULL,	/* 20h: 16-bit 64k code at 0x00000000 */
	0x000092000000ffffULL	/* 28h: 16-bit 64k data at 0x00000000 */
};

static struct {
	unsigned short size __attribute__ ((packed));
	unsigned long long *base __attribute__ ((packed));
} real_mode_gdt = {
sizeof(real_mode_gdt_entries) - 1, 0}

, real_mode_idt = {
0x3ff, 0};

/*
  Registers:
  eax - scratch
  ebx - memory list pointer
  ecx - counter
  edx - entry point
  esi - scratch memcpy pointer
  edi - scratch memcpy pointer
  ebp - flags
 */
/* The address arguments to this function are PHYSICAL ADDRESSES */
static void
real_mode_switch(struct monte_reloc_page_t *memonte_list,
		 void *entry, unsigned long flags)
{
	__asm__ __volatile__(
		/*
		 * Grab the args for our memcpy off the stack now while we still
		 * have page tables.
		 */
		"    movl %0, %%ebx           \n"	/* Arg storage: memlist -> EBX */
		"    movl %1, %%edx           \n"	/* Arg storage: entry   -> EDX */
		"    movl %2, %%ebp           \n"	/* Arg storage: flags   -> EBP */
			
		/*
		 * Turn off paging, leave protected mode turned on.
		 */
		"    movl %%cr0, %%eax        \n"	/* Turn off paging (bit 31 in CR0) */
		"    andl $0x7FFFFFFF, %%eax  \n"
		"    movl %%eax, %%cr0        \n" "    xorl %%eax, %%eax        \n"	/* Flush the TLB (write 0 to CR3) */
		"    movl %%eax, %%cr3        \n"

		/*
		 * Memory setup code
		 *
		 * We need to memcpy all the bits of code and data reload to
		 * the right places now We can do this safely now that we don't
		 * have to worry about overwriting our own page tables.
		 */
		"    mov  $0x18, %%ax         \n"	/* 32-bit, 4GB data from GDT */
		"    mov  %%ax, %%ds          \n"
		"    mov  %%ax, %%es          \n"
		"    cld                      \n"
		"1:                           \n" "    orl  %%ebx, %%ebx        \n"	/* ebx = addr of chunk descriptor */
		"    jz   3f                  \n" "    movl  (%%ebx), %%edi     \n"	/* Destination */
		"    movl 4(%%ebx), %%esi     \n"	/* Source */
		"    movl $0x1000, %%ecx      \n"	/* Size = PAGE_SIZE */
		"    cmpl %%edi, %%esi        \n"	/* skip if dest == src  */
		"    je   2f                  \n" "    rep; movsb               \n"	/* Do it. */
		"2:                           \n" "    addl $8, %%ebx           \n"	/* move ebx to the next item. */
		"    movl %%ebx, %%edi        \n"	/* Check if we're at the end    */
		"    andl $0xfff, %%edi       \n"	/* of the page and actually     */
		"    cmpl $0xff8, %%edi       \n"	/* looking at the next page ptr. */
		"    jne  1b                  \n"	/* If not, next iteration */
		"        movl (%%ebx), %%ebx  \n"	/* Follow the pointer here */
		"    jmp  1b                  \n"	/* ... next iteration */
		"3:                           \n"
		
		/*
		 * Option here...  If we want to skip the setup section, stay
		 * in protected mode and jump to the new kernel now.
		 *
		 * bit 0 of EBP: 1 = stay in protected mode start; skip the setup
		 *                   portion of the kernel.  (Jump to the kernel
		 *                   from here.)
		 *               0 = continue on to real mode; start kernel at
		 *                   setup code.
		 */
		"    andl $0x1, %%ebp         \n"
		"    jz  1f                   \n" "        xor %%bx, %%bx       \n"	/* bx==0 means boot */
		"        jmp *%%edx           \n"	/* Jump to our entry point */
		"1:                           \n"
		
		/*
		 * Now that our memcpy is done we can get to 16 bit code
		 * segment.  This configures CS properly for real mode.
		 */
		"    ljmp $0x20, $0x1000-(real_mode_switch_end - __rms_16bit) \n"
		"__rms_16bit:                 \n" ".code16                      \n"	/* 16 bit code from here on... */

		/*
		 * Load the segment registers w/ properly configured segment
		 * descriptors.  They will retain these configurations (limits,
		 * writability, etc.) once protected mode is turned off.
		 */
		"    mov  $0x28, %%ax         \n"
		"    mov  %%ax, %%ds          \n"
		"    mov  %%ax, %%es          \n"
		"    mov  %%ax, %%fs          \n"
		"    mov  %%ax, %%gs          \n"
		"    mov  %%ax, %%ss          \n"

		/*
		 * Turn off protection (bit 0 in CR0)
		 */
		"    movl %%cr0, %%eax        \n"
		"    andl $0xFFFFFFFE, %%eax  \n"
		"    movl %%eax, %%cr0        \n"

		/*
		 * Now really going into real mode
		 */
		"    ljmp $0, $0x1000-(real_mode_switch_end - __rms_real) \n"
		"__rms_real:                  \n"

		/*
		 * Setup a stack
		 */
		"    mov  $0x9000, %%ax       \n"
		"    mov  %%ax, %%ss          \n"
		"    mov  $0xAFFE, %%ax       \n"
		"    mov  %%ax, %%sp          \n"

		/*
		 * Dump zeros in the other segregs
		 */
		"    xor  %%ax, %%ax          \n"
		"    mov  %%ax, %%ds          \n"
		"    mov  %%ax, %%es          \n"
		"    mov  %%ax, %%fs          \n"
		"    mov  %%ax, %%gs          \n" "    sti                      \n"	/* Enable interrupts */

		/*
		 * Try and sanitize the video hardware state.
		 */
		"    mov  $0x0003, %%ax       \n"	/* Ask for 80x25 */
		"    int  $0x10               \n" "    push %%edx               \n"	/* Kludge to do far jump */
		"    lret                     \n"
		/*"    jmp *%%edx               \n" */
		/*"    ljmp $0x9020, $0x0000    \n" */
#if 0
		/* Debugging tools... */
		"    .byte 0xcc               \n"	/* XXX DEBUG triple fault XXXX */
		"    .byte 0xeb, 0xfe         \n"	/* XXX DEBUG wedge XXXXX */
#endif
		".code32                      \n"	/* Restore mode for rest of file */
		::"m"(memonte_list), "m"(entry), "m"(flags));
}

__asm__(".text\n" "real_mode_switch_end:\n");
extern char real_mode_switch_end[];

/*
 * Functions to turn off PCI bus masters.  This doesn't seem to be a
 * good thing to do in practice.  Unloading drivers before hand is a
 * better mechanism to turn things off.
 */
#ifdef CONFIG_PCI

/*
 * This needs three things to work:
 *   1) a kernel compiled with CONFIG_HOTPLUG
 *   2) an audit to be sure all compiled in pci driver have their
 *      unload methods declared __devexit and not __exit
 *      (alternative: use only modular pci drivers)
 *   3) a little kernel patch to export the variable
 *      pci_drivers in drivers/pci/pci.c (linux-2.4.0-pci.diff)
 *
 * Do NOT enable this unconditionally, you will fail.  -- hch
 */
static void
monte_pci_unload_drivers(void)
{
#ifdef __UNLOAD_PCI_DRIVERS__
	struct list_head *l;

	list_for_each(l, pci_drivers)
		pci_unregister_driver(list_entry(l, struct pci_driver, node));

#endif /* __UNLOAD_PCI_DRIVERS__ */
}

static void
monte_pci_disable_bus(struct pci_bus *bus)
{
	struct pci_dev *dev;
	struct list_head *l;
	u16 cmd;

	for (l = bus->devices.next; l != &bus->devices; l = l->next) {
		dev = pci_dev_b(l);
		pcibios_read_config_word(dev->bus->number, dev->devfn,
					 PCI_COMMAND, &cmd);
		cmd &= ~PCI_COMMAND_MASTER;
		pcibios_write_config_word(dev->bus->number, dev->devfn,
					  PCI_COMMAND, cmd);
	}
}

static void
monte_pci_disable(void)
{
	struct list_head *l;
	
	/*
	 * Turn off PCI bus masters to keep them from scribbling
	 * on our memory later on.
	 */
	if (!pcibios_present())
		return;
	monte_pci_unload_drivers();
	for (l = pci_root_buses.next; l != &pci_root_buses; l = l->next)
		monte_pci_disable_bus(pci_bus_b(l));
}
#endif				/* CONFIG_PCI */

static void restore_xt_pic(void)
{
	/* These following is taken from arch/i386/boot/setup.S

	 *     I hope. Now we have to reprogram the interrupts :-( we put
	 *     them right after the intel-reserved hardware interrupts, at
	 *     int 0x20-0x2F. There they won't mess up anything. Sadly IBM
	 *     really messed this up with the original PC, and they
	 *     haven't been able to rectify it afterwards. Thus the bios
	 *     puts interrupts at 0x08-0x0f, which is used for the
	 *     internal hardware interrupts as well. We just have to
	 *     reprogram the 8259's, and it isn't fun.
	 *
	 * This tidbit puts the interrupts back where the BIOS likes them. */

	outb(0x11, 0x20);	/* 8259A-1 Start initialization sequence */
	outb(0x11, 0xA0);	/* 8259A-2 Start initialization sequence */

	outb(0x08, 0x21);	/* 1: start hw interrupts at 0x08 */
	outb(0x0F, 0xA1);	/* 2: start hw interrupts at 0x0F */

	outb(0x04, 0x21);	/* 1: this PIC is master */
	outb(0x02, 0xA1);	/* 2: this PIC is slave */

	outb(0x01, 0x21);	/* 1: 8086 mode */
	outb(0x01, 0xA1);	/* 2: 8086 mode */

	outb(0x00, 0xA1);	/* unmask all interrupts */
	outb(0x00, 0x21);
}

static int monte_restart(unsigned long entry_addr, unsigned long flags)
{
	void *ptr;
	cli();

	/*
	 * Step 1: get an identity mapped page.  The code that will turn
	 * off paging later needs to run out of an identity mapped page.
	 * For simplicity we'll use page zero.  This page is normally not
	 * mapped at all.
	 */
	set_bit(PG_reserved, &(virt_to_page(__va(0)))->flags);
	if (remap_page_range(0, 0, PAGE_SIZE, PAGE_KERNEL)) {
		clear_bit(PG_reserved, &(virt_to_page(__va(0)))->flags);
		return -EAGAIN;
	}

	/*----- POINT OF NO RETURN IS HERE --------------------------------------*/

	/*
	 * Step 2: Copy the real_mode_switch code and our GDT into the
	 * page we just setup.  We stick this stuff right at the end to
	 * avoid clobbering BIOS variables and stuff in the beginning.
	 */
#define RMS_SIZE ((unsigned long)real_mode_switch_end - (unsigned long) real_mode_switch)
	/* Copy our GDT down there. */
	ptr = (void *)
		((0x1000 - RMS_SIZE - sizeof(real_mode_gdt_entries)) & ~7);
	memcpy(ptr, real_mode_gdt_entries, sizeof(real_mode_gdt_entries));
	real_mode_gdt.base = ptr;

	/* Copy the real_mode_switch function down there. */
	memcpy((void *) (0x1000 - RMS_SIZE), real_mode_switch, RMS_SIZE);

	monte_make_phys();		/* Prep our memory to-do list. */

	/* Prep the hardware to restart */
#ifdef CONFIG_PCI
	if (flags & MONTE_PCI_DISABLE)
		monte_pci_disable();	/* Turn off PCI bus masters */
#endif
	if (!(flags & MONTE_PROTECTED))
		restore_xt_pic();

	/*
	 * Ok, now the real monkey business begins.... Please keep hands
	 * and feet inside the memory space and remain seated until the
	 * ride comes to a complete stop.
	 */
	__asm__ __volatile__(
		/*
		 * Install the IDT and GDT we copied to page zero.
		 */

		"lidt  %0             \n"
		"lgdt  %1             \n"
		/* Function call with args... sort of */
		"pushl %4             \n"
		"pushl %3             \n" "pushl %2             \n"	/* Push args on the stack */
		"sub $4, %%esp        \n"	/* bogo return address */
		"ljmp $0x10, $0x1000-(real_mode_switch_end-real_mode_switch) \n"::
		"m"(real_mode_idt), "m"(real_mode_gdt),
		"r"(monte_pg_list), "r"(entry_addr),
		"r"(flags):"memory");

	/* NOT REACHED */

	while (1)
		/* Shut up gcc. */;
}


/*
 *  Module mechanics
 */
extern void *sys_call_table[];

static int __init
monte_init(void)
{
	real_reboot = sys_call_table[__NR_reboot];
	sys_call_table[__NR_reboot] = sys_monte;
	return 0;
}

static void __exit
monte_exit(void)
{
	sys_call_table[__NR_reboot] = real_reboot;
}

module_init(monte_init);
module_exit(monte_exit);
